@@ -6,8 +6,8 @@ The main application run loop.
 
 from __future__ import print_function
 from __future__ import unicode_literals
-
-__version__ = "0.0.6"
+from __future__ import absolute_import
+from __future__ import division
 
 import os
 import sys
@@ -15,16 +15,19 @@ if os.name == 'nt':
     print('sermon is not compatabile with Windows.')
     sys.exit()
 
-import argparse
-import glob
 import threading
 import re
 import time
+import argparse
 
 import serial
-from serial.tools import list_ports
 import urwid
 
+import sermon
+import sermon.util as util
+from sermon.magics import magic
+from sermon.resources import help_status_str
+
 try:
     input = raw_input
 except NameError:
@@ -61,10 +64,10 @@ class ConsoleEdit(urwid.Edit):
             # go.
             if self.history_pos == len(self.history):
                 # Already as far as we can go in history.
-                beep()
+                util.beep()
                 return
-            self.history_pos = limit(self.history_pos + 1,
-                                     1, len(self.history))
+            self.history_pos = util.limit(self.history_pos + 1,
+                                          1, len(self.history))
             self.set_edit_text(
                 self.history[-self.history_pos])
             self.set_edit_pos(len(self.edit_text))
@@ -74,11 +77,11 @@ class ConsoleEdit(urwid.Edit):
                 self.history_pos = 0
                 return False
             elif self.history_pos == 0:
-                beep()
+                util.beep()
                 return
 
-            self.history_pos = limit(self.history_pos - 1,
-                                     1, len(self.history))
+            self.history_pos = util.limit(self.history_pos - 1,
+                                          1, len(self.history))
             self.set_edit_text(
                 self.history[-self.history_pos])
             self.set_edit_pos(len(self.edit_text))
@@ -87,7 +90,41 @@ class ConsoleEdit(urwid.Edit):
             return super(ConsoleEdit, self).keypress(size, key)
 
 
-class Sermon():
+class ScrollingTextOverlay(urwid.Overlay):
+    def __init__(self, content, bottom_widget):
+        """
+        Parameters
+        ----------
+        content : str
+            The contents to display in the overlay widget.
+        bottom_widget : urwid.Widget
+            The original widget that the overlay appears over.
+        """
+        listbox = urwid.ListBox([urwid.Text(content)])
+        frame = urwid.Frame(listbox,
+                            header=urwid.AttrMap(
+                                urwid.Text(help_status_str), 'statusbar'))
+
+        super(ScrollingTextOverlay, self).__init__(
+            frame, bottom_widget,
+            align='center', width=('relative', 100),
+            valign='top', height=('relative', 100),
+            left=0, right=0,
+            top=0, bottom=0)
+
+    def keypress(self, size, key):
+        if key == 'j':
+            key = 'down'
+        elif key == 'k':
+            key = 'up'
+        elif key == 'ctrl d':
+            key = 'page down'
+        elif key == 'ctrl u':
+            key = 'page up'
+        return super(ScrollingTextOverlay, self).keypress(size, key)
+
+
+class Sermon(object):
     """
     The main serial monitor class. Starts a read thread that polls the serial
     device and prints results to top window. Sends commands to serial device
@@ -95,23 +132,35 @@ class Sermon():
     """
     def __init__(self, device, args):
         # Receive display widgets
-        self.receive_text = urwid.Text('')
-        body = urwid.ListBox([self.receive_text, urwid.Text('')])
+        self.receive_window = urwid.Text('')
+        body = urwid.ListBox([self.receive_window, urwid.Text('')])
         body.set_focus(1)
 
-        # Send edit widgets
-        footer = ConsoleEdit(self.on_edit_done, ': ')
+        # Draw main frame with status header and footer for commands.
+        self.conection_msg = urwid.Text('', 'left')
+        self.status_msg = urwid.Text('', 'right')
+        self.header = urwid.Columns([self.conection_msg, self.status_msg])
+        self.frame = urwid.Frame(
+            body,
+            header=urwid.AttrMap(self.header, 'statusbar'),
+            footer=ConsoleEdit(self.on_edit_done, ': '),
+            focus_part='footer')
+        palette = [
+            ('error', 'light red', 'black'),
+            ('ok', 'dark green', 'black'),
+            ('statusbar', '', 'black')
+        ]
+        self.loop = urwid.MainLoop(self.frame, palette,
+                                   unhandled_input=self.unhandled_key_handler)
 
-        self.frame = urwid.Frame(body, footer=footer, focus_part='footer')
-        self.loop = urwid.MainLoop(self.frame)
-
-        self.fd = self.loop.watch_pipe(self.received_output)
+        self.fd = self.loop.watch_pipe(self.received_data)
 
         self.kill = False
         self.append_text = args.append.encode(
             'latin1').decode('unicode_escape')
         self.frame_text = args.frame.encode('latin1').decode('unicode_escape')
-        self.byte_list_pattern = re.compile('(\$\(([^\)]+)\))')
+        self.byte_list_pattern = re.compile(
+            '(\$\(([^\)]+?)\))|(\${([^\)]+?)})')
         self.device = device
         self.serial = serial.Serial(device,
                                     baudrate=args.baud,
@@ -124,16 +173,81 @@ class Sermon():
                                     timeout=0.1)
         time.sleep(0.1)
         self.serial.flushInput()
+        self.conection_msg.set_text(('ok', self.serial.name))
 
         self.worker = threading.Thread(target=self.serial_read_worker)
         self.worker.daemon = True
 
-    def on_edit_done(self, edit_text):
-        self.write_command(edit_text)
+        self.logging = False
+        self.logfile = None
+        magic.app = self
 
-    def received_output(self, data):
-        self.receive_text.set_text(self.receive_text.text +
-                                   data.decode('latin1'))
+    def update_status(self, status, text):
+        self.status_msg.set_text((status, text))
+
+    def unhandled_key_handler(self, key):
+        if key in ('q', 'esc'):
+            self.loop.widget = self.frame
+
+    def on_edit_done(self, edit_text):
+        """
+        Callback called when editing is completed (after enter is pressed)
+        """
+        if len(edit_text) < 1:
+            self.update_status('ok', '')
+            return
+        if edit_text[0] == '%':
+            # Handle magic command.
+            try:
+                result = magic.execute(edit_text[1:])
+                if result['status'] is not None:
+                    self.update_status('ok', result['status'])
+                if result['bytes_to_send'] is not None:
+                    self.serial.write(result['bytes_to_send'])
+            except (util.ArgumentParseError, AttributeError, ValueError) as e:
+                self.update_status('error', str(e))
+            return
+        if self.logging:
+            self.update_status('ok', 'Logging to %s' % self.logfile)
+        processed_command = ('%(frame)s%(command)s%(append)s%(frame)s' %
+                             {'frame': self.frame_text,
+                              'append': self.append_text,
+                              'command': edit_text})
+        # matches list of bytes $(0x08, 0x09, ... )
+        strings = self.byte_list_pattern.split(processed_command)
+        n = 0
+        while n < len(strings) - 1:
+            if strings[n] is None or strings[n] == '':
+                n += 1
+            elif self.byte_list_pattern.match(strings[n]):
+                # list of bytes
+                self.write_list_of_bytes(strings[n+1])
+                n += 2
+            else:
+                self.serial.write(strings[n].encode('latin1'))
+                n += 1
+        self.serial.write(strings[-1].encode('latin1'))
+
+    def overlay(self, content):
+        """
+        Shows the given content in an overlay window above the main frame.
+
+        Parameters
+        ----------
+        content : str
+            The text content to display over the main frame.
+        """
+        self.loop.widget = ScrollingTextOverlay(content, self.frame)
+
+    def received_data(self, data):
+        self.receive_window.set_text(self.receive_window.text +
+                                     data.decode('latin1'))
+        if self.logging:
+            try:
+                with open(self.logfile, 'a') as f:
+                    f.write(data.decode('latin1'))
+            except:
+                self.update_status('error', 'Error writing to logfile.')
 
     def serial_read_worker(self):
         """
@@ -156,33 +270,8 @@ class Sermon():
                     raise
 
     def write_list_of_bytes(self, string):
-        split_str = [s.strip() for s in string.split(',')]
-        for s in split_str:
-            try:
-                if sys.version_info.major == 3:
-                    self.serial.write(chr(int(s, 0) & 255).encode('latin1'))
-                else:
-                    self.serial.write(chr(int(s, 0) & 255))
-            except:
-                pass
-
-    def write_command(self, command):
-        processed_command = ('%(frame)s%(command)s%(append)s%(frame)s' %
-                             {'frame': self.frame_text,
-                              'append': self.append_text,
-                              'command': command})
-        # matches list of bytes $(0x08, 0x09, ... )
-        strings = self.byte_list_pattern.split(processed_command)
-        n = 0
-        while n < len(strings) - 1:
-            if self.byte_list_pattern.match(strings[n]):
-                # list of bytes
-                self.write_list_of_bytes(strings[n+1])
-                n += 2
-            else:
-                self.serial.write(strings[n].encode('latin1'))
-                n += 1
-        self.serial.write(strings[-1].encode('latin1'))
+        byte_data = [int(s.strip(), 0) for s in string.split(',')]
+        self.serial.write(bytearray(byte_data))
 
     def start(self):
         self.worker.start()
@@ -194,43 +283,9 @@ class Sermon():
             pass
         self.serial.close()
 
-
-def beep():
-    sys.stdout.write('\a')
-
-
-def limit(n, minimum, maximum):
-    """
-    Limits n to be within minimum and maximum.
-    """
-    if n < minimum:
-        return minimum
-    elif n > maximum:
-        return maximum
-    return n
-
-
-def serial_devices():
-    """
-    Returns a list of the available serial devices.
-    """
-    if sys.platform == 'darwin' and sys.version_info.major == 3:
-        # pyserial's builtin port detection not working on mac with python 3
-        return glob.glob('/dev/cu.*')
-    else:
-        return [p[0] for p in list_ports.comports()]
-
-
-def print_serial_devices():
-    """
-    Prints all of the available serial devices.
-    """
-    devices = serial_devices()
-    if len(devices) == 0:
-        print('No serial devices found.')
-        return
-    for p in devices:
-        print(p)
+    def exit(self):
+        self.stop()
+        raise urwid.ExitMainLoop()
 
 
 def main():
@@ -288,17 +343,17 @@ def main():
 
     # List serial devices and exit for argument '-l'
     if commandline_args.list:
-        print_serial_devices()
+        util.print_serial_devices()
         sys.exit()
     elif commandline_args.version:
-        print(__version__)
+        print(sermon.__version__)
         sys.exit()
 
     # If device is not specified, prompt user to select an available device.
     device = None
     if not commandline_args.device:
         try:
-            devices = serial_devices()
+            devices = util.serial_devices()
             if len(devices) > 0:
                 print('')
                 for n in range(len(devices)):
@@ -323,12 +378,12 @@ def main():
     commandline_args.stopbits = stopbits_values[commandline_args.stopbits]
 
     try:
-        sermon = Sermon(device, commandline_args)
+        app = Sermon(device, commandline_args)
     except serial.serialutil.SerialException as e:
-        print(str(e))
+        print(e)
         sys.exit(1)
 
     try:
-        sermon.start()
+        app.start()
     except KeyboardInterrupt:
-        sermon.stop()
+        app.stop()
